package org.processmining.logprojection;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.extension.XExtension;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeContinuous;
import org.deckfour.xes.model.XAttributeDiscrete;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.framework.plugin.Progress;
import org.processmining.log.utils.XUtils;
import org.processmining.logprojection.calculators.AbstractCategoryCalculator.ContinuousCategoryCalculator;
import org.processmining.logprojection.calculators.AbstractCategoryCalculator.DiscreteCategoryCalculator;
import org.processmining.logprojection.sorting.TraceSorter;

import gnu.trove.map.TObjectIntMap;
import gnu.trove.map.hash.TObjectIntHashMap;

public class LogView implements DataManager.StringHandler {

	private static final long serialVersionUID = 5922420564904570094L;

	protected static final char[] UNKNOWN = "UNKNOWN by LogView".toCharArray();

	protected final XEventDataManager eventManager;
	protected final XTraceDataManager traceManager;
	protected final XLogDataManager logManager;

	private int literalBlocks = 1;
	private final int litBlockSize = 1024;
	private char[][][] index2lit = new char[1][litBlockSize][];
	private int literals = 0;

	private int[] traces;

	private int numberOfEvents;

	private int numberOfTraces;

	private HashMap<String, XExtension> key2extension;

	private ArrayList<XEventClassifier> classifiers;

	public LogView(XLog log) {
		this(log, null);
	}

	public LogView(XLog log, Progress progress) {
		TObjectIntMap<String> lit2index = new TObjectIntHashMap<String>(1024, 0.5f, -1);
		addLiteral(new String(DataManager.StringHandler.UNKNOWN), lit2index);

		Set<XExtension> exts = log.getExtensions();
		this.key2extension = new HashMap<String, XExtension>(log.getExtensions().size());
		for (XExtension extension : exts) {
			for (XAttribute attribute : extension.getDefinedAttributes()) {
				this.key2extension.put(attribute.getKey(), extension);
			}
		}

		classifiers = new ArrayList<XEventClassifier>(log.getClassifiers());

		Map<String, XAttribute> eventAttributeMap = new HashMap<String, XAttribute>();
		Map<String, XAttribute> traceAttributeMap = new HashMap<String, XAttribute>();

		int t = 0;
		if (progress != null) {
			progress.setMaximum(2 * log.size() + 1);
		}
		boolean eventsProcessed = false;
		for (t = 0; t < log.size()
				&& !(eventsProcessed && eventAttributeMap.isEmpty() && traceAttributeMap.isEmpty()); t++) {
			if (progress != null) {
				if (progress.isCancelled()) {
					break;
				}
				progress.inc();
			}
			XTrace trace = log.get(t);
			if (t == 0) {
				Iterator<XAttribute> it = trace.getAttributes().values().iterator();
				while (it.hasNext()) {
					XAttribute att = it.next();
					if (!traceAttributeMap.containsKey(att.getKey())) {
						traceAttributeMap.put(att.getKey(), att);
					}
				}
			} else {
				Iterator<String> it = traceAttributeMap.keySet().iterator();
				while (it.hasNext()) {
					String att = it.next();
					if (!trace.getAttributes().containsKey(att) || //
					// Extremely ugly way to check if type of attribute is equal to type of 
					// previously encountered attribute. Currently it's not possible to ask for
					// the type of an attribute. FIXME
							!XUtils.isSameType(traceAttributeMap.get(att), trace.getAttributes().get(att))) {
						//							!traceAttributeMap.get(att).getClass()
						//									.isAssignableFrom(trace.getAttributes().get(att).getClass())) {
						System.err.println("Trace attribute not global: " + att + " in trace " + t);
						if (progress != null) {
							progress.setCaption("Trace attribute not global: " + att + " in trace " + t);
						}
						it.remove();
					}
				}
			}
			for (int e = 0; e < trace.size() && !(eventsProcessed && eventAttributeMap.isEmpty()); e++) {
				if (progress != null) {
					if (progress.isCancelled()) {
						break;
					}
				}
				XEvent event = trace.get(e);
				if (!eventsProcessed) {
					Iterator<XAttribute> it = event.getAttributes().values().iterator();
					while (it.hasNext()) {
						XAttribute att = it.next();
						if (!eventAttributeMap.containsKey(att.getKey())) {
							eventAttributeMap.put(att.getKey(), att);
						}
					}
					//					eventAttributes.addAll(trace.get(e).getAttributes().values());
					eventsProcessed = true;
				} else {
					Iterator<String> it = eventAttributeMap.keySet().iterator();
					while (it.hasNext()) {
						String att = it.next();
						if (!event.getAttributes().containsKey(att) || //
						// Extremely ugly way to check if type of attribute is equal to type of 
						// previously encountered attribute. Currently it's not possible to ask for
						// the type of an attribute. FIXME
								!XUtils.isSameType(eventAttributeMap.get(att), event.getAttributes().get(att))) {
							//								!eventAttributeMap.get(att).getClass()
							//										.isAssignableFrom(event.getAttributes().get(att).getClass())) {
							System.err.println(
									"Event attribute not global: " + att + " in event " + e + " of trace " + t);
							if (progress != null) {
								progress.setCaption(
										"Event attribute not global: " + att + " in event " + e + " of trace " + t);
							}
							it.remove();
						}
					}
				}
			}

		}
		if (progress != null)

		{
			progress.setValue(log.size());
		}
		for (XAttribute att : log.getGlobalEventAttributes()) {
			if (!eventAttributeMap.containsKey(att.getKey())) {
				eventAttributeMap.put(att.getKey(), att);
			} else {
				// compare types
				// types are stored in the class hierarchy. Ugly as hell!
				if (!eventAttributeMap.get(att.getKey()).getClass().isAssignableFrom(att.getClass())) {
					eventAttributeMap.remove(att.getKey());
					if (progress != null) {
						progress.setCaption(
								"Event Attribute type of global " + att.getKey() + " does not match it usage.");
					}
					System.err.println("Event Attribute type of global " + att.getKey() + " does not match it usage.");
				}
			}
		}
		for (XAttribute att : log.getGlobalTraceAttributes()) {
			if (!traceAttributeMap.containsKey(att.getKey())) {
				traceAttributeMap.put(att.getKey(), att);
			} else {
				// compare types
				// types are stored in the class hierarchy. Ugly as hell!
				if (!traceAttributeMap.get(att.getKey()).getClass().isAssignableFrom(att.getClass())) {
					traceAttributeMap.remove(att.getKey());
					if (progress != null) {
						progress.setCaption(
								"Trace Attribute type of global " + att.getKey() + " does not match it usage.");
					}
					System.err.println("Trace Attribute type of global " + att.getKey() + " does not match it usage.");
				}
			}
		}

		if (eventAttributeMap.isEmpty() && traceAttributeMap.isEmpty()) {
			throw new RuntimeException("Cannot load a log without global attributes");
		}

		eventManager = new XEventDataManager(key2extension, this, eventAttributeMap.values(), lit2index);
		traceManager = new XTraceDataManager(key2extension, this, traceAttributeMap.values(), lit2index);
		logManager = new XLogDataManager(key2extension, this, log.getAttributes().values(), lit2index);

		Set<String> errorKeys = new HashSet<String>();
		Set<String> eventErrorKeys = new HashSet<String>();
		logManager.addObject(log, lit2index, errorKeys);
		numberOfEvents = 0;
		traces = new int[log.size()];
		t = 0;
		for (Iterator<String> it = errorKeys.iterator(); it.hasNext();) {
			System.err.println("Log Attribute not global:   " + it.next());
		}
		errorKeys.clear();
		for (XTrace trace : log) {

			if (progress != null) {
				if (progress.isCancelled()) {
					break;
				}
				progress.inc();
			}

			traceManager.addObject(trace, lit2index, errorKeys, trace.size());

			int e = 0;
			for (XEvent event : trace) {

				if (progress != null) {
					if (progress.isCancelled()) {
						break;
					}
				}

				if (e == 0) {
					traces[t] = numberOfEvents;
				}

				eventManager.addObject(event, lit2index, eventErrorKeys, t, e);

				numberOfEvents++;
				e++;
			}
			t++;

		}
		numberOfTraces = traces.length;

		for (Iterator<String> it = errorKeys.iterator(); it.hasNext();) {
			String s = it.next();
			if (progress != null) {
				progress.setCaption("Trace Attribute not global: " + s);
			}
			System.err.println("Trace Attribute not global: " + s);

		}
		for (Iterator<String> it = eventErrorKeys.iterator(); it.hasNext();) {
			String s = it.next();
			if (progress != null) {
				progress.setCaption("Event Attribute not global: " + s);
			}
			System.err.println("Event Attribute not global: " + s);
		}

		for (String key : eventAttributeMap.keySet()) {
			if (eventAttributeMap.get(key) instanceof XAttributeContinuous) {
				ContinuousCategoryCalculator<XEvent> calculator = new ContinuousCategoryCalculator<>(this,
						eventManager.getAttributeForKey(key), true);
				eventManager.addAttributes(calculator, calculator.getAttributes());
			}
			if (eventAttributeMap.get(key) instanceof XAttributeDiscrete) {
				DiscreteCategoryCalculator<XEvent> calculator = new DiscreteCategoryCalculator<>(this,
						eventManager.getAttributeForKey(key), true);
				eventManager.addAttributes(calculator, calculator.getAttributes());
			}
		}
		for (String key : traceAttributeMap.keySet()) {
			if (traceAttributeMap.get(key) instanceof XAttributeContinuous) {
				ContinuousCategoryCalculator<XTrace> calculator = new ContinuousCategoryCalculator<>(this,
						traceManager.getAttributeForKey(key), false);
				traceManager.addAttributes(calculator, calculator.getAttributes());
			}
			if (traceAttributeMap.get(key) instanceof XAttributeDiscrete) {
				DiscreteCategoryCalculator<XTrace> calculator = new DiscreteCategoryCalculator<>(this,
						traceManager.getAttributeForKey(key), false);
				traceManager.addAttributes(calculator, calculator.getAttributes());
			}
		}
	}

	public LogView(Collection<XAttribute> globalEventAttributes, Collection<XAttribute> globalTraceAttributes,
			TObjectIntMap<String> lit2index) {

		addLiteral(new String(DataManager.StringHandler.UNKNOWN), lit2index);

		this.key2extension = new HashMap<String, XExtension>(1);

		classifiers = new ArrayList<XEventClassifier>(1);

		eventManager = new XEventDataManager(key2extension, this, globalEventAttributes, lit2index);
		traceManager = new XTraceDataManager(key2extension, this, globalTraceAttributes, lit2index);
		logManager = new XLogDataManager(key2extension, this, Collections.<XAttribute>emptySet(), lit2index);

		numberOfEvents = 0;
		traces = new int[0];
		numberOfTraces = traces.length;
	}

	public void addEvent(XEvent event, TObjectIntMap<String> lit2index, Set<String> errorKeysEvents,
			int... headerData) {
		eventManager.addObject(event, lit2index, errorKeysEvents, headerData);
		numberOfEvents++;
	}

	public void addTrace(XTrace trace, TObjectIntMap<String> lit2index, Set<String> errorKeysEvents,
			Set<String> errorKeysTraces) {
		traceManager.addObject(trace, lit2index, errorKeysTraces, trace.size());
		traces = Arrays.copyOfRange(traces, 0, traces.length + 1);
		traces[numberOfTraces] = numberOfEvents;
		int e = 0;
		for (XEvent event : trace) {
			addEvent(event, lit2index, errorKeysEvents, numberOfTraces, e);
			e++;
		}
		numberOfTraces++;
	}

	public Map<String, XExtension> getExtensions() {
		return key2extension;
	}

	@Override
	public String getLiteral(int index) {
		return new String(getLiteralChars(index));
	}

	@Override
	public char[] getLiteralChars(int index) {
		int block = index / litBlockSize;
		int idx = index % litBlockSize;
		return index2lit[block][idx];
	}

	@Override
	public synchronized int addLiteral(String value, TObjectIntMap<String> lit2index) {

		int exists = lit2index.get(value);

		if (exists < 0) {
			if (literals == literalBlocks * litBlockSize) {
				if (literalBlocks == index2lit.length) {
					// need to resize
					if (literalBlocks > Integer.MAX_VALUE / 2) {
						throw new RuntimeException("Storage Full");
					}
					// double the storage.
					char[][][] oldStore = index2lit;
					index2lit = new char[2 * literalBlocks][][];
					System.arraycopy(oldStore, 0, index2lit, 0, literalBlocks);
				}
				// we need to add a block
				index2lit[literalBlocks] = new char[litBlockSize][];
				literalBlocks++;
			}

			int block = literals / litBlockSize;
			int index = literals % litBlockSize;
			index2lit[block][index] = value.toCharArray();
			lit2index.put(value, literals);
			return literals++;
		} else {
			return exists;
		}

	}

	public XEvent readEvent(int trace, int event) {
		return readEvent(traces[trace] + event);
	}

	public XEvent readEvent(int number) {
		return eventManager.readObject(number);
	}

	public XTrace readEmptyTrace(int trace) {
		return traceManager.readObject(trace);
	}

	public XTrace readTrace(int trace) {
		XTrace t = readEmptyTrace(trace);
		int size = traceManager.getTraceSize(trace);
		for (int i = traces[trace]; i < traces[trace] + size; i++) {
			t.add(readEvent(i));
		}
		return t;
	}

	public XEventDataManager getEventDataManager() {
		return eventManager;
	}

	public XTraceDataManager getTraceDataManager() {
		return traceManager;
	}

	public int getFirstEventNumber(int traceNumber) {
		return traces[traceNumber];
	}

	public int numTraces() {
		return numberOfTraces;
	}

	public int numEvents() {
		return numberOfEvents;
	}

	public int getTraceLength(int trace) {
		return traceManager.getTraceSize(trace);
	}

	public int getTraceForEvent(int event) {
		return eventManager.getHeader(event, 0);
	}

	public synchronized void sortTraces(TraceSorter comparator) {
		// create current sort;
		Integer[] objects = new Integer[numberOfTraces];
		for (int i = 0; i < objects.length; i++) {
			objects[i] = i;
		}
		Arrays.sort(objects, comparator);
		// now swap the data.
		for (int i = 0; i < numberOfTraces; i++) {
			// we need to swap object at index [i] with index [objects[i]]
			int loc = objects[i];
			while (loc < i) {
				loc = objects[loc];
			}
			if (i == loc) {
				continue;
			}
			int oldEventStart = traces[loc];
			// swap the trace data in the trace manager
			traceManager.swap(i, loc);
			// swap the locations of the first events for the traces
			traces[loc] = traces[i];
			traces[i] = oldEventStart;
			// all events that were previously pointing to trace "loc"
			// should now point to trace i;
			for (int e = 0; e < getTraceLength(i); e++) {
				eventManager.updateHeader(traces[i] + e, i, e);
			}
			// all events that were previously pointing to trace "i"
			// should now point to trace loc;
			for (int e = 0; e < getTraceLength(loc); e++) {
				eventManager.updateHeader(traces[loc] + e, loc, e);
			}
		}
	}

	public StatisticsWithSpan getEventStatisticsDiscrete(Attribute attribute) {

		StatisticsWithSpan stat = new StatisticsWithSpan.Long();

		long min = Long.MAX_VALUE;
		long max = Long.MIN_VALUE;
		int lastTrace = -1;
		int d = 0;
		for (int b = 0; b < eventManager.blocks; b++) {
			int offset = eventManager.key2index.get(attribute);
			for (int e = 0; d < numberOfEvents
					&& e < eventManager.blockSize; e += eventManager.bytesReservedPerObject) {
				int t = eventManager.readInt(eventManager.store[b], e);
				if (t != lastTrace) {
					// new trace;
					if (lastTrace >= 0) {
						stat.addSpanObservation(max - min);
					}
					min = Long.MAX_VALUE;
					max = Long.MIN_VALUE;
					lastTrace = t;
				}
				long val = eventManager.readLong(eventManager.store[b], e + offset);
				if (val < min) {
					min = val;
				}
				if (val > max) {
					max = val;
				}
				stat.addObservation(val);
				d++;
			}
		}
		return stat;
	}

	public StatisticsWithSpan getEventStatisticsContinuous(Attribute attribute) {
		StatisticsWithSpan stat = new StatisticsWithSpan.Double();
		double min = Double.MAX_VALUE;
		double max = Double.MIN_VALUE;
		int lastTrace = -1;
		int d = 0;
		for (int b = 0; b < eventManager.blocks; b++) {
			int offset = eventManager.key2index.get(attribute);
			for (int e = 0; d < numberOfEvents
					&& e < eventManager.blockSize; e += eventManager.bytesReservedPerObject) {
				int t = eventManager.readInt(eventManager.store[b], e);
				if (t != lastTrace) {
					// new trace;
					if (lastTrace >= 0) {
						stat.addSpanObservation(max - min);
					}
					min = Double.MAX_VALUE;
					max = Double.MIN_VALUE;
					lastTrace = t;
				}
				double val = eventManager.readDouble(eventManager.store[b], e + offset);
				if (val < min) {
					min = val;
				}
				if (val > max) {
					max = val;
				}
				stat.addObservation(val);
				d++;
			}
		}
		return stat;
	}

	public StatisticsWithSpan getEventStatisticsTimestamp(Attribute attribute) {
		return getEventStatisticsDiscrete(attribute);
	}

	public Statistics getTraceStatisticsDiscrete(Attribute attribute) {
		return traceManager.getStatisticsDiscrete(attribute);
	}

	public Statistics getTraceStatisticsContinuous(Attribute attribute) {
		return traceManager.getStatisticsContinuous(attribute);
	}

	public Statistics getTraceStatisticsTimestamp(Attribute attribute) {
		return traceManager.getStatisticsTimestamp(attribute);
	}

	public Statistics getStatisticsDiscrete(Attribute attribute) {
		if (attribute.isEventAttribute()) {
			return getEventStatisticsDiscrete(attribute);
		} else {
			return traceManager.getStatisticsDiscrete(attribute);
		}
	}

	public Statistics getStatisticsContinuous(Attribute attribute) {
		if (attribute.isEventAttribute()) {
			return getEventStatisticsContinuous(attribute);
		} else {
			return traceManager.getStatisticsContinuous(attribute);
		}
	}

	public int getDifferentValueCount(Attribute a) {
		if (a.isEventAttribute()) {
			return eventManager.getAttributeValuesLiteral(a, false).length;
		} else {
			return traceManager.getAttributeValuesLiteral(a, false).length;
		}
	}

	public Statistics getStatisticsTimestamp(Attribute attribute) {
		if (attribute.isEventAttribute()) {
			return getEventStatisticsDiscrete(attribute);
		} else {
			return traceManager.getStatisticsTimestamp(attribute);
		}
	}

	public XLogDataManager getLogDataManager() {
		return logManager;
	}

	public List<XEventClassifier> getClassifiers() {
		return classifiers;
	}

	public Histogram getHistogramDiscrete(Attribute attribute, Statistics stats, int buckets) {
		if (attribute.isEventAttribute()) {
			return eventManager.getHistogramDiscrete(attribute, (Statistics.Long) stats, buckets);
		} else {
			return traceManager.getHistogramDiscrete(attribute, (Statistics.Long) stats, buckets);
		}
	}

	public Histogram getHistogramContinuous(Attribute attribute, Statistics stats, int buckets) {
		if (attribute.isEventAttribute()) {
			return eventManager.getHistogramContinuous(attribute, (Statistics.Double) stats, buckets);
		} else {
			return traceManager.getHistogramContinuous(attribute, (Statistics.Double) stats, buckets);
		}
	}

	public Histogram getHistogramTimestamp(Attribute attribute, Statistics stats, int buckets) {
		if (attribute.isEventAttribute()) {
			return eventManager.getHistogramTimestamp(attribute, (Statistics.Long) stats, buckets);
		} else {
			return traceManager.getHistogramTimestamp(attribute, (Statistics.Long) stats, buckets);
		}
	}

	public void writeToFile(File file) throws FileNotFoundException, IOException {
		ObjectOutputStream out = new ObjectOutputStream(new BufferedOutputStream(new FileOutputStream(file)));
		try {
			out.writeObject(this);
		} finally {
			out.close();
		}
	}

	public static LogView readFromFile(File file) throws ClassNotFoundException, IOException {
		ObjectInputStream in = new ObjectInputStream(new BufferedInputStream(new FileInputStream(file)));
		try {
			return (LogView) in.readObject();
		} finally {
			in.close();
		}
	}

}
