package org.processmining.logprojection.abstraction;

import java.util.Arrays;

import org.processmining.logprojection.abstraction.util.TIntArrayIntHashMap;
import org.processmining.logprojection.lightxlog.LightLog;
import org.processmining.logprojection.lightxlog.LightTrace;

public class DirectSuccessionCounter implements CausalDependencyMatrix {

	public final static int NOEVENT = -1;

	private final String[] int2EventClass;
	private final int[] occcount;
	private final int[][] succount;
	private final int[][] abacount;
	private final LightLog log;

	private final TIntArrayIntHashMap featureSets = new TIntArrayIntHashMap();

	public DirectSuccessionCounter(LightLog log) {
		this.log = log;
		this.int2EventClass = new String[log.getInt2EventClass().length + 2];
		for (int i = 1; i < int2EventClass.length - 1; i++) {
			int2EventClass[i] = log.getInt2EventClass()[i - 1];
		}
		int2EventClass[0] = new String("Start");
		int2EventClass[int2EventClass.length - 1] = new String("End");

		occcount = new int[log.getNumEventClasses() + 2];
		succount = new int[log.getNumEventClasses() + 2][log
				.getNumEventClasses() + 2];
		abacount = new int[log.getNumEventClasses() + 2][log
				.getNumEventClasses() + 2];

		int size = 4;
		int[] window = new int[size + 1];
		Arrays.fill(window, NOEVENT);
		for (int t = 0; t < log.size(); t++) {
			LightTrace trace = log.getTrace(t);

			if (trace.size() == 0) {
				continue;
			}

			int a = 0;
			// artificial first event
			occcount[a]++;
			window = Arrays.copyOfRange(window, 1, window.length + 1);
			window[size] = a;
			featureSets.adjustOrPutValue(window, 1, 1);

			int b = trace.getEvent(0).getNameIndex() + 1;
			occcount[b]++;
			succount[0][b]++;

			if (trace.size() > 1) {
				// shift the elements of the window left by one index
				window = Arrays.copyOfRange(window, 1, window.length + 1);
				window[size] = b;
				featureSets.adjustOrPutValue(window, 1, 1);

				int c = 0;
				for (int i = 1; i < trace.size(); i++) {
					c = trace.getEvent(i).getNameIndex() + 1;

					// shift the elements of the window left by one index
					window = Arrays.copyOfRange(window, 1, window.length + 1);
					window[size] = c;
					featureSets.adjustOrPutValue(window, 1, 1);

					occcount[c]++;
					if (c == a && c != b) {
						abacount[a][b]++;
					}
					succount[b][c]++;
					a = b;
					b = c;
				}
			}
			a = b;
			b = log.getNumEventClasses() + 1;
			// artificial last event
			occcount[b]++;
			succount[a][b]++;
			window = Arrays.copyOfRange(window, 1, window.length + 1);
			window[size] = b;

			for (int i = 0; i < size; i++) {
				// shift the elements of the window left by one index
				window = Arrays.copyOfRange(window, 1, window.length + 1);
				window[size] = NOEVENT;
				featureSets.adjustOrPutValue(window, 1, 1);
			}
		}
	}

	public boolean isCausal(int from, int to) {
		return (succount[from][to] > 0 && (succount[to][from] == 0 || from == to))
				|| (abacount[from][to] + abacount[to][from] > 0);
	}

	public boolean isParallel(int from, int to) {
		return from != to && (succount[from][to] > 0 && succount[to][from] > 0)
				&& (abacount[from][to] + abacount[to][from] == 0);
	}

	public boolean isUnrelated(int from, int to) {
		return (succount[from][to] == 0 && succount[to][from] == 0);
	}

	public String getEventClass(int index) {
		return int2EventClass[index];
	}

	public int getNumEventClasses() {
		return int2EventClass.length;
	}

	public boolean isSucceeding(int from, int to) {
		return succount[from][to] > 0;
	}

	public int getOccurenceCount(int eventClass) {
		return occcount[eventClass];
	}

	public LightLog getLog() {
		return log;
	}

	public int getSucCount(int i, int j) {
		return succount[i][j];
	}

	public int getABACount(int i, int j) {
		return abacount[i][j];
	}

	@Override
	public double getCausalDependency(int node1, int node2) {
		return succount[node1][node2]
				/ (double) Math.min(occcount[node1], occcount[node2]);
	}

	@Override
	public int getNumberOfNodes() {
		return int2EventClass.length;
	}

	public String toString() {
		return "Direct Succession";
	}

	@Override
	public int getCount(int node1, int node2) {
		return succount[node1][node2];
	}

}
