package org.processmining.logprojection.plugins.dottedchart.filtering;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JTabbedPane;
import javax.swing.WindowConstants;

import org.processmining.logprojection.Attribute;
import org.processmining.logprojection.filters.EventFilter;
import org.processmining.logprojection.plugins.dottedchart.DottedChart;

public class AttributeFilterDialog extends JDialog implements WindowListener,
		ActionListener {

	private static final long serialVersionUID = -2426901678654855237L;
	private final List<AttributeFilterPanel> panels = new ArrayList<AttributeFilterPanel>();
	private final DottedChart chart;
	private JButton applyButton;
	private JButton closeButton;

	public AttributeFilterDialog(JFrame owner, DottedChart chart) {
		this(owner, chart, new HashMap<Attribute, int[]>());
	}

	public AttributeFilterDialog(JFrame owner, DottedChart chart,
			Map<Attribute, int[]> attribute2selected) {
		super(owner, "Select attribute...", true);
		this.chart = chart;

		JTabbedPane tabbedPane = new JTabbedPane(JTabbedPane.TOP);

		JTabbedPane eventPane = setupFilterTabs(chart,
				AttributeFilterPanel.Type.EVENT);
		JTabbedPane firstEventPane = setupFilterTabs(chart,
				AttributeFilterPanel.Type.FIRSTEVENTINTRACE);
		JTabbedPane lastEventPane = setupFilterTabs(chart,
				AttributeFilterPanel.Type.LASTEVENTINTRACE);
		JTabbedPane anyEventPane = setupFilterTabs(chart,
				AttributeFilterPanel.Type.ANYEVENTINTRACE);

		tabbedPane.addTab("Event Filters", eventPane);
		tabbedPane.addTab("First event filters", firstEventPane);
		tabbedPane.addTab("Last event filters", lastEventPane);
		tabbedPane.addTab("Event presence filters", anyEventPane);

		this.add(tabbedPane, BorderLayout.CENTER);

		Dimension d = new Dimension(800, 600);
		this.setMinimumSize(d);
		this.setSize(d);

		this.setDefaultCloseOperation(WindowConstants.HIDE_ON_CLOSE);

		final Dimension screenSize = Toolkit.getDefaultToolkit()
				.getScreenSize();
		final int x = (screenSize.width - getWidth()) / 2;
		final int y = (screenSize.height - getHeight()) / 2;
		setLocation(x, y);

		this.addWindowListener(this);

		JPanel buttonPanel = new JPanel(new FlowLayout(FlowLayout.RIGHT));

		applyButton = new JButton("Apply");
		applyButton.addActionListener(this);
		buttonPanel.add(applyButton);

		closeButton = new JButton("OK");
		closeButton.addActionListener(this);
		buttonPanel.add(closeButton);

		this.add(buttonPanel, BorderLayout.SOUTH);

		this.pack();

	}

	private JTabbedPane setupFilterTabs(DottedChart chart,
			AttributeFilterPanel.Type type) {
		JTabbedPane eventPane = new JTabbedPane(JTabbedPane.TOP);

		LiteralAttributeFilterPanel literalFilterPanel = new LiteralAttributeFilterPanel(
				chart.getLogView(), type);
		if (literalFilterPanel.getAttributeCount() > 0) {
			panels.add(literalFilterPanel);
			eventPane.addTab("Literal", literalFilterPanel);
		}
		BooleanAttributeFilterPanel booleanFilterPanel = new BooleanAttributeFilterPanel(
				chart.getLogView(), type);
		if (booleanFilterPanel.getAttributeCount() > 0) {
			panels.add(booleanFilterPanel);
			eventPane.addTab("Boolean", booleanFilterPanel);
		}

		ContinuousAttributeFilterPanel continuousFilterPanel = new ContinuousAttributeFilterPanel(
				chart.getLogView(), type);
		if (continuousFilterPanel.getAttributeCount() > 0) {
			panels.add(continuousFilterPanel);
			eventPane.addTab("Continuous", continuousFilterPanel);
		}
		DiscreteAttributeFilterPanel discreteFilterPanel = new DiscreteAttributeFilterPanel(
				chart.getLogView(), type);
		if (discreteFilterPanel.getAttributeCount() > 0) {
			panels.add(discreteFilterPanel);
			eventPane.addTab("Discrete", discreteFilterPanel);
		}
		AbsoluteTimestampAttributeFilterPanel timestampAbsFilterPanel = new AbsoluteTimestampAttributeFilterPanel(
				chart.getLogView(), type);
		if (timestampAbsFilterPanel.getAttributeCount() > 0) {
			panels.add(timestampAbsFilterPanel);
			eventPane.addTab("Timestamp", timestampAbsFilterPanel);
		}
		RelativeTimestampAttributeFilterPanel timestampRelFilterPanel = new RelativeTimestampAttributeFilterPanel(
				chart.getLogView(), type);
		if (timestampRelFilterPanel.getAttributeCount() > 0) {
			panels.add(timestampRelFilterPanel);
			eventPane.addTab("Relative time", timestampRelFilterPanel);
		}
		return eventPane;
	}

	@Override
	public void windowOpened(WindowEvent e) {
		for (AttributeFilterPanel panel : panels) {
			panel.opening();
		}
	}

	@Override
	public void windowClosing(WindowEvent e) {
		// for (AttributeFilterPanel panel : panels) {
		// doFiltering(panel);
		// }
	}

	@Override
	public void windowClosed(WindowEvent e) {

	}

	@Override
	public void windowIconified(WindowEvent e) {
	}

	@Override
	public void windowDeiconified(WindowEvent e) {
	}

	@Override
	public void windowActivated(WindowEvent e) {
	}

	@Override
	public void windowDeactivated(WindowEvent e) {
		for (AttributeFilterPanel panel : panels) {
			doFiltering(panel);
		}
	}

	@Override
	public void actionPerformed(ActionEvent e) {
		if (e.getSource() == closeButton) {
			this.setVisible(false);
		} else {
			for (AttributeFilterPanel panel : panels) {
				doFiltering(panel);
			}

		}
	}

	private void doFiltering(AttributeFilterPanel panel) {
		panel.closing();

		boolean addNew = !panel.getFiltersToAdd().isEmpty();

		Iterator<EventFilter> it = panel.getFiltersToRemove().iterator();
		while (it.hasNext()) {
			chart.removeEventFilter(it.next(), !it.hasNext() && !addNew);
		}

		it = panel.getFiltersToAdd().iterator();
		while (it.hasNext()) {
			chart.addEventFilter(it.next(), !it.hasNext());
		}
	}
}
